#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>

int main(int argc, char *argv[])
{
  struct addrinfo hints, *gai, *ai;
  int err;
  int yes = 1;
  int srv, clt;
  void *addr;
  in_port_t port;
  char ip[INET6_ADDRSTRLEN];
  struct sockaddr_storage from;
  int len = sizeof(from);
  uint8_t buf[1024];

  if (argc != 2) {
    fprintf(stderr, "Usage: %s PORT\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  /* binding server */
  bzero(&hints, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;
  if ((err = getaddrinfo(NULL, argv[1], &hints, &gai)) != 0) {
    fprintf(stderr, "getaddrinfo: port %s on localhost: %s\n", argv[1], gai_strerror(err));
    exit(EXIT_FAILURE);
  }
  for (ai = gai; ai != NULL; ai = ai->ai_next) {
    if ((srv = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol)) < 0) {
      perror("socket");
      continue;
    }
    if (setsockopt(srv, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) < 0) {
      perror("setsockopt");
      exit(EXIT_FAILURE);
    }
    if (bind(srv, ai->ai_addr, ai->ai_addrlen) < 0) {
      close(srv);
      perror("bind");
      continue;
    }
    break;
  }
  freeaddrinfo(gai);
  if (ai == NULL) {
    fprintf(stderr, "Failed to bind server.\n");
    exit(EXIT_FAILURE);
  }

  /* waiting for client */
  if (listen(srv, 8) < 0) {
    perror("listen");
    close(srv);
    exit(EXIT_FAILURE);
  }
  addr = (ai->ai_family == AF_INET)
    ? (void *)&(((struct sockaddr_in *)ai->ai_addr)->sin_addr)
    : (void *)&(((struct sockaddr_in6 *)ai->ai_addr)->sin6_addr);
  inet_ntop(ai->ai_family, addr, ip, sizeof(ip));
  port = (ai->ai_family == AF_INET)
    ? ((struct sockaddr_in *)ai->ai_addr)->sin_port
    : ((struct sockaddr_in6 *)ai->ai_addr)->sin6_port;
  printf("> Server listening on %s:%d…\n", ip, ntohs(port));

  if ((clt = accept(srv, (struct sockaddr *)&from, &len)) < 0) {
    perror("accept");
    close(srv);
    exit(EXIT_FAILURE);
  }
  addr = (from.ss_family == AF_INET)
    ? (void *)&(((struct sockaddr_in *)&from)->sin_addr)
    : (void *)&(((struct sockaddr_in6 *)&from)->sin6_addr);
  inet_ntop(from.ss_family, addr, ip, sizeof(ip));
  port = (from.ss_family == AF_INET)
    ? ((struct sockaddr_in *)&from)->sin_port
    : ((struct sockaddr_in6 *)&from)->sin6_port;
  printf("> Connection from %s:%d\n", ip, ntohs(port));


  /* echoing */
  for (;;) {
    bzero(buf, sizeof(buf));
    if ((len = recv(clt, buf, sizeof(buf), 0)) < 0) {
      perror("read");
      continue;
    }
    printf("> Received %d bytes.\n", len);
    if (len == 0) {
      printf("> Closing session.\n");
      break;
    }
    if ((len = send(clt, buf, len, 0)) < 0) {
      perror("write");
    }
    printf("> Send %d bytes back.\n", len);
  }

  close(clt);
  close(srv);

  return 0;
}
